/*	com.c

	AFAPI communication function definitions for SHMAPERS.

	July 1996 by H. Dietz
*/


#include <string.h>
#include "afapi.h"
#include "inline.h"

/*	p_bcastPut/Get() for all data types...
*/

#define	P_DECLARE(suffix) \
void \
p_bcastPut##suffix(register p_##suffix d) \
{ \
	/* Different from most op because it uses PE NPROC buffer */ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	q->datum[ which ].u_##suffix = d; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) return; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) return; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
} \
 \
p_##suffix \
p_bcastGet##suffix(void) \
{ \
	/* Different from most op because it uses PE NPROC buffer */ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Signal I am at the barrier */ \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Return broadcast value... */ \
	return(P_CAST_##suffix(q->datum[ which ].u_##suffix)); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_bcastPutz() and p_bcastGetz()...

	Well, compression isn't really effective for SHMAPERS, so we
	just send the block a few bytes per synchronization....  This
	code makes use of the fact that 386, 486, and Pentium
	processors don't mind making unaligned memory accesses.
*/

void
p_bcastPutz(register void *source,
register int bytes)
{
	while (bytes > 0) {
		register int size = ((bytes > P_BLOCKSIZE) ?
				     P_BLOCKSIZE :
				     bytes);
		register void *p = ((void *) &(p_shm_all[NPROC + 1]));
		if (p_shm_me->barno & 1) p += P_BLOCKSIZE;

		/* Copy the block... */
		memcpy(((void *) p), source, size);

		/* Sync */
		p__wait();

		/* Next... */
		source += size;
		bytes -= size;
	}
}


void
p_bcastGetz(register void *dest,
register int bytes)
{
	while (bytes > 0) {
		register int size = ((bytes > P_BLOCKSIZE) ?
				     P_BLOCKSIZE :
				     bytes);
		register void *p = ((void *) &(p_shm_all[NPROC + 1]));
		if (p_shm_me->barno & 1) p += P_BLOCKSIZE;

		/* Sync */
		p__wait();

		/* Copy the block... */
		memcpy(dest, ((void *) p), size);

		/* Next... */
		dest += size;
		bytes -= size;
	}
}

/*	p_putGet() functions...

	This is very close to what the SHMAPERS mechanism does
	naturally, and is thus quite efficient.
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_putGet##suffix(register p_##suffix datum, \
register int source) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = datum; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Return the one value we want... */ \
	return(P_CAST_##suffix( \
		p_shm_all[ source ].datum[ which ].u_##suffix)); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_gather() functions...

	This is very close to what SHMAPERS does naturally.  However,
	we still have to copy the natural gather results from their
	shared memory locations into the destination specified.
*/

#define	P_DECLARE(suffix) \
void \
p_gather##suffix(register p_##suffix *dest, \
register p_##suffix datum) \
{ \
	register volatile p_shm_type *p = p_shm_all; \
	register volatile p_shm_type *q = (p + NPROC); \
	register int barno = p_shm_me->barno + 1; \
	register int which = (barno & 1); \
	register int barnonext = (barno + 1); \
 \
	/* Install my datum and then signal barrier */ \
	p_shm_me->datum[ which ].u_##suffix = datum; \
	p_shm_me->barno = barno; \
 \
	/* Now wait for everybody else... */ \
	do { \
		register int hisbarno = p->barno; \
		if (hisbarno == barnonext) goto syncd; \
		if (hisbarno != barno) { \
			do { \
				P_SHM_IDLE; \
				hisbarno = p->barno; \
				if (hisbarno == barnonext) goto syncd; \
			} while (hisbarno != barno); \
		} \
	} while (++p < q); \
 \
syncd: \
	/* Copy the values into dest... */ \
	p = p_shm_all; \
	do { \
		*dest = P_CAST_##suffix(p->datum[ which ].u_##suffix); \
		++dest; \
	} while (++p < q); \
}

P_ALPHA

#undef	P_DECLARE

/*	p_scatPut() and p_scatGet() functions...

	This is really a very simple operation to construct using the
	other routines, but it has been added to the library by
	popular demand....  It is not very efficient.
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scatPut##suffix(register p_##suffix *src) \
{ \
	register int who = 0; \
 \
	do { \
		p_bcastPut##suffix(P_CAST_##suffix(src[who])); \
	} while (++who < NPROC); \
 \
	return(src[IPROC]); \
} \
 \
p_##suffix \
p_scatGet##suffix(void) \
{ \
	register p_##suffix t; \
	register p_##suffix result = 0; \
	register int who = 0; \
 \
	do { \
		t = p_bcastGet##suffix(); \
		if (who == IPROC) result = t; \
	} while (++who < NPROC); \
 \
	return(result); \
}

P_ALPHA

#undef	P_DECLARE
